/*
 * e2extend: Prepares an freshly mke2fs file system for being
 * elegible to be added to an existing ext2-volume
 *
 * TODO:
 *   - Allow for multiple devices to be extended instead of
 *     forcing the user to do it one at a time.
 * 
 * Miguel de Icaza (miguel@nuclecu.unam.mx), 1997.
 */
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <linux/ext2_fs.h>
#include <ext2fs/ext2fs.h>

/* To fetch the conversion routines */
#define __KERNEL__
#include <asm/byteorder.h>

/* Number of devices we support */
#define MAXSUPER               128

/* Trigger to detect a non-fresh fs */
#define MAX_INODES_ON_FRESH_FS 4

typedef struct {
	int  fd;
	char *name;
	long super_block;
	struct ext2_super_block es;

	int blocks;
	int block_groups;
} sb_info_t;

static sb_info_t super_blocks [MAXSUPER];
static int sb_count;
static long total_blocks;
static ext2_filsys fs = NULL;
int retval;

void
usage (void)
{
	fprintf (stderr, "usage is:\n"
		 "e2extend [[-b superblock] [-f devname]...]\n\n"
		 "where the last pair is the new device\n");
}

int
report_ext2_err (char *msg)
{
	com_err("e2extend", retval, "While %s\n", msg);
	
	return 1;
}

int
main (int argc, char *argv [])
{
	sb_info_t *new;
	struct ext2_inode inode;
	struct ext2_group_desc *s;
	ext2_inode_scan   scan;
	void *block_buf;
	ino_t ino;
	int c, i;
	int total_inodes = 0;
	int goof = 0;

	while ((c = getopt (argc, argv, "b:f:")) != EOF){
		switch (c){
		case 'b':
			super_blocks [sb_count].super_block = atoi (optarg);
			break;
		case 'f':
			super_blocks [sb_count].name = optarg;
			if (super_blocks [sb_count].super_block == 0)
				super_blocks [sb_count].super_block = 1;
			sb_count++;
			break;
		default:
			usage ();
		}
	}

	if (sb_count < 2){
		usage ();
		return 1;
	}
	for (i = 0; i < sb_count-1; i++){
		sb_info_t *sb = &super_blocks [i];
		int fd;
		
		if ((fd = open (sb->name, O_RDONLY)) == -1){
			perror (sb->name);
			return 1;
		}
		sb->fd = fd;
		if (lseek (fd, sb->super_block * 1024, SEEK_SET) == -1){
			perror (sb->name);
			return 1;
		}
		if (read (fd, &sb->es, sizeof (sb->es)) == -1){
			perror (sb->name);
			return 1;
		}
		sb->blocks = le32_to_cpu (sb->es.s_blocks_count);
		printf ("Device %s has %ld blocks\n", sb->name, sb->blocks);
		total_blocks += sb->blocks;

		close (fd);
	}

	new = &super_blocks [i];

	/* For now, ignore the super_block offset, let ext2fs lib handle it */
	printf ("Extension device [%s] first_block = %ld\n", new->name, total_blocks);
	retval = ext2fs_open (new->name, EXT2_FLAG_RW, 0, 0, unix_io_manager, &fs);
	if (retval)
		return report_ext2_err ("opening");

	retval = ext2fs_read_bitmaps (fs);
	if (retval)
		return report_ext2_err ("loading bitmaps");

	block_buf = (void *) malloc (fs->blocksize * 3);
	if (block_buf == NULL){
		perror ("no more memory");
		return 1;
	}

	/* 1. Check the sanity of the soon to be opened file system */
	retval = ext2fs_open_inode_scan (fs, 0, &scan);
	if (retval)
		return report_ext2_err ("openning inode scan");

	retval = ext2fs_get_next_inode(scan, &ino, &inode);
	if (retval)
		return report_ext2_err ("starting the inode scan");

	/* 1.1. Scan the inodes: check if there are too many of them,
	 *      or if they have too many data blocks
	 */
	for (total_inodes = 0; ino;) {
		if (inode.i_blocks){
			total_inodes++;
			if (inode.i_block [EXT2_IND_BLOCK]){
				printf ("inode %d has indirect blocks not good (%d)\n",
					ino, inode.i_block [EXT2_IND_BLOCK]);
				goof++;
			}
			printf ("Inode %ld has %d blocks\n", ino);
		}
		retval = ext2fs_get_next_inode(scan, &ino, &inode);
		if (retval)
			return report_ext2_err ("fetching the next inode");
	}
	ext2fs_close_inode_scan (scan);

	if (total_inodes > MAX_INODES_ON_FRESH_FS){
		printf ("Mhm, too many inodes for an extension fs: %d/%d\n",
			MAX_INODES_ON_FRESH_FS, total_inodes);
		goof++;
	}
	
	if (goof)
		return 1;

	printf ("Remapping %d inodes\n", total_inodes);
	
	/* 2. Make the inodes point to the the locations with a global pointer */
	retval = ext2fs_open_inode_scan (fs, 0, &scan);
	if (retval)
		return report_ext2_err ("openning inode scan");

	retval = ext2fs_get_next_inode(scan, &ino, &inode);
	if (retval)
		return report_ext2_err ("starting the inode scan");
	
	for (total_inodes = 0; ino;) {
		if (inode.i_blocks){
			unsigned int o_block = inode.i_block [0];
			inode.i_block [0] += total_blocks;
			printf ("Inode %d block %ld -> %ld\n", ino, o_block, inode.i_block [0]);
			retval = ext2fs_write_inode (fs, ino, &inode);
			if (retval)
				return report_ext2_err ("recording the inode change\n");
		}
		retval = ext2fs_get_next_inode(scan, &ino, &inode);
		if (retval)
			return report_ext2_err ("fetching the next inode");
	}
	ext2fs_close_inode_scan (scan);

	ext2fs_close (fs);
	
	/* For now, ignore the super_block offset, let ext2fs lib handle it */
	printf ("Extension device [%s] first_block = %ld\n", new->name, total_blocks);
	retval = ext2fs_open (new->name, EXT2_FLAG_RW, 0, 0, unix_io_manager, &fs);
	if (retval)
		return report_ext2_err ("opening");

	retval = ext2fs_read_bitmaps (fs);
	if (retval)
		return report_ext2_err ("loading bitmaps");

	for (i = 0, s=fs->group_desc; i < fs->group_desc_count; i++){
		printf ("group %d: incremented in %d blocks\n", i, total_blocks);
		s->bg_block_bitmap += total_blocks;
		s->bg_inode_bitmap += total_blocks;
		s->bg_inode_table  += total_blocks;
		s++;
	}
	retval = ext2fs_flush (fs);
	if (retval)
		return report_ext2_err ("updating new file system block group descriptor offsets");
		
	return 0;
}
